#include "setcalc.h"
#include <stdio.h>
#include <stdlib.h>

void shellSort(long *item, short count)  {

    /* these are going to be very small arrays coming in - so shell sort
     * is an OK algorithm i guess.  since they're so small, we only need 3
     * numbers in the 'a' array, the size of the incoming set, 1/2 that, 
     * and 1 (required for shell sort)
     */

    register short i, j, gap, k;
    short x, a[3];

    a[0] = count;
    a[1] = count / 2;
    a[2] = 1;


    for(k=0; k<3; k++)  {
        gap = a[k];
        for (i = gap; i < count; ++i)  {
            x = item[i];
            for (j = i - gap; x < item[j] && j >= 0; j = j - gap) {
                item[j+gap] = item[j];
	    }
            item[j+gap] = x;
        }
    }
    
}

short enforceRange(long *item, short count)  {
        /*  we need to enforce the fact that pitch classes can only
        be in the range of 0 to 11.  */
        
    register short i;
    for(i = 0; i < count; i++)  {
        if(item[i] > 11 || item[i] < 0)  {
            item[i] = item[count-1];
            --i; --count;
        }
    }
    return count;
}


short removeDuplicates(long *item, short count)  {
    
    /* our set range here is finite, so we can just set up a tally for the 12 pitches (0-11) */
    short tally[12];
    register short i,x;

    /* set all tallys to false initially */
    for(i = 0; i < 12; i++)  {
        tally[i] = 0;
    }	

    for(i = 0; i < count; i++)  {
        x = item[i];	
	    if(tally[x] == 0)  {
            tally[x] = 1;
	    }
	    else {
	        /* if we find a dupe, swap it with the end of the array,
	         * decrement the number of elements, and rewind the loop
	         */
	         item[i] = item[count-1];
	         item[count-1] = x;
	         --i; --count; 
	    }
    }
    return count; 
}


short generateCompliment(long *item, short count, long *compliment) {
    
    /*  generate the list of complimentary pitch classes */
    short tally[12];
    register short i, x, compliment_loc;
    
    /* set all tallys to false initially */
    for(i = 0; i < 12; i++)  {
        tally[i] = 0;
    }
    
    for (i = 0; i < count; i++)  {
        x = item[i];
        if (tally[x] == 0)  {
            tally[x] = 1;
        }
    }
    
    compliment_loc = 0;
    for (i = 0; i < 12; i++)  {
        if (tally[i] == 0)  {
            compliment[compliment_loc] = i;
            compliment_loc++;
        }
    }
    
    return compliment_loc;  // this is a count of how many items are now in the compliment array
    
}


void findNormalForm(long *item, short count, long *normal)  {

    long matrix[12][12];
    register short i,j,k,smallest;
    short min, minIndex = 0;
    long dist, first[12];
    long duplicates[12][12];
    short duplicateCount = 0;
    short rowCount = count;
    short colCount = count;
    // int shortSize = sizeof(short);
    // short rowSize = shortSize*count;

    
    /*  create a matrix of all possible orderings of this set */
    for (i = 0; i < count; i++)  {
        for (j = 0; j < count; j++)  {
            matrix[i][j] = item[j];
        }
        k = item[0];
        for(j = 0; j < count - 1; j++)  {
            item[j] = item[j+1];
        }
        item[count - 1] = k;
    }
    
    
    /*  algorithm for finding the normal form */
    while(1)  {
        min = 12;
        for (i = 0; i < rowCount; i++) {
            dist = matrix[i][colCount - 1] - matrix[i][0];
            if (dist < 0) {
                dist += 12;
            }

            if (dist < min) { /* we have a new minimum */
                duplicateCount = 0;
                min = dist;
                minIndex = i;
                memcpy(duplicates[duplicateCount++], matrix[i], sizeof(matrix[i]));
                continue;
            }
            if (dist == min) {
                memcpy(duplicates[duplicateCount++], matrix[i], sizeof(matrix[i]));
            }
        }


        if (duplicateCount == 0) {  /* this means the array has length 1, so there is only 1 min */
            break;
        }
        if (duplicateCount == count - 1)  {  /* symmetrical set */
            break;
        }

        rowCount = duplicateCount;
        duplicateCount = 0;
        j = sizeof(duplicates);
        memcpy(matrix, duplicates, j);
        colCount--;

    }
    
    
    if (duplicateCount != 0)  {
        /* find smallest first value and that is our normal */
        for(i = 0; i < count; i++)  {
            first[i] = matrix[i][0];
        }
        smallest = returnSmallest(first, count);
    }
    else  {
        /* minIndex is our normal */
        smallest = minIndex;
    }


    /* populate the normal form array with the correct row */
    for(i = 0; i < count; i++)  {
        normal[i] = matrix[smallest][i];
    }
    
    return;

}


void findPrimeForm(long *item, short count, long *prime)  {
    
    register short i, normalTotal = 0, invertedTotal = 0, dist;
    long inverted[12];

    /* transpose the set so the first element is 0 */
    dist = item[0];
    for(i = 0; i < count; i++)  {
	    prime[i] = item[i] - dist;
	    if (prime[i] < 0)  {
            prime[i] += 12;
	    }
	    normalTotal += prime[i];
    }

    /* invert the set */
    for(i = 0; i < count; i++)  {
	    inverted[i] = 0 - (item[i]);
	    if (inverted[i] < 0)  {
            inverted[i] += 12;
	    }
    }

    /* find the normal form of the inverted set */
    shellSort(inverted, count);
    findNormalForm(inverted, count, inverted);

    /* transpose the inverted set to start on 0 */
    dist = inverted[0];
    for(i = 0; i < count; i++)  {
        inverted[i] = inverted[i] - dist;
        if (inverted[i] < 0)  {
            inverted[i] += 12;
        }
	invertedTotal += inverted[i];
    }

    
    /* figure out which one (the inverted or the regular) set
     * is more packed to the left.  i think we can do this by
     * simply adding up the members.  I hope that works.
     * anyway, we've been keeping track of the total above.
     */

    if (invertedTotal < normalTotal)  {
        memcpy(prime, inverted, sizeof(inverted));
    }
    else {
	    /*  nothing, we're fine with prime the way it is
	     */
    }
    return;
}



short returnSmallest(long *list, short count)  {
    
    /* this function will return the index of the smallest
     * element in the given array, -1 if there are any ties
     */
     
    long i,tiny,loc,tim;
    loc = 0;
    tiny = list[0];
    tim = 13; /* out of range */

    for(i = 1; i < count; i++)  {
       if(list[i] < tiny)  {
           loc = i;
       }
       if(list[i] == tiny)  {
	   tim = list[i];
       }
    }       

    if (tiny == tim)  {
        return -1;
    }

    return loc;

}





void getForteNumber(long *prime, short count, char *forte)  {

    /*  return the forte number for the given prime form */

    /*  someone remind me to be nice to Dr. Forte if i ever meet him */

    /* we can cut down on the search time for forte numbers
     * given the total members in the set.  since our
     * possibilities are finite, we can predefine
     * our search field now, and only have to do
     * a linear search on up to 50 elements
     * instead of 200+
     */

    int members = 0;
    int zsets = 0;
    register int i = 0;
    register int j;
    int f2 = 0; /* second part of the forte number */
    int z = 0; /* flag for Z-sets */
    char theZ;

    int threes[12][3] = {{0,1,2}, {0,1,3}, {0,1,4}, {0,1,5}, 
	                     {0,1,6}, {0,2,4}, {0,2,5}, {0,2,6}, 
			             {0,2,7}, {0,3,6}, {0,3,7} ,{0,4,8}};

    int fours[29][4] =  {{0,1,2,3}, {0,1,2,4}, {0,1,3,4}, {0,1,2,5},
	                     {0,1,2,6}, {0,1,2,7}, {0,1,4,5}, {0,1,5,6}, 
			             {0,1,6,7}, {0,2,3,5}, {0,1,3,5}, {0,2,3,6}, 
			             {0,1,3,6}, {0,2,3,7}, {0,1,4,6}, {0,1,5,7}, 
			             {0,3,4,7}, {0,1,4,7}, {0,1,4,8}, {0,1,5,8}, 
			             {0,2,4,6}, {0,2,4,7}, {0,2,5,7}, {0,2,4,8}, 
			             {0,2,6,8}, {0,3,5,8}, {0,2,5,8}, {0,3,6,9}, 
			             {0,1,3,7}};
    int foursZ[2] = {15, 29};

    int fives[38][5] = {{0,1,2,3,4}, {0,1,2,3,5}, {0,1,2,4,5}, {0,1,2,3,6}, 
	                    {0,1,2,3,7}, {0,1,2,5,6}, {0,1,2,6,7}, {0,2,3,4,6}, 
			            {0,1,2,4,6}, {0,1,3,4,6}, {0,2,3,4,7}, {0,1,3,5,6}, 
			            {0,1,2,4,8}, {0,1,2,5,7}, {0,1,2,6,8}, {0,1,3,4,7}, 
			            {0,1,3,4,8}, {0,1,4,5,7}, {0,1,3,6,7}, {0,1,5,6,8}, 
			            {0,1,4,5,8}, {0,1,4,7,8}, {0,2,3,5,7}, {0,1,3,5,7}, 
			            {0,2,3,5,8}, {0,2,4,5,8}, {0,1,3,5,8}, {0,2,3,6,8},
		       	        {0,1,3,6,8}, {0,1,4,6,8}, {0,1,3,6,9}, {0,1,4,6,9},
		       	        {0,2,4,6,8}, {0,2,4,6,9}, {0,2,4,7,9}, {0,1,2,4,7},
		                {0,3,4,5,8}, {0,1,2,5,8}}; 
    int fivesZ[6] = {36,38,17,12,18,37};  

    int sixes[50][6] = {{0,1,2,3,4,5}, {0,1,2,3,4,6}, {0,1,2,3,5,6}, 
	                    {0,1,2,4,5,6}, {0,1,2,3,6,7}, {0,1,2,5,6,7}, 
			            {0,1,2,6,7,8}, {0,2,3,4,5,7}, {0,1,2,3,5,7}, 
			            {0,1,3,4,5,7}, {0,1,2,4,5,7}, {0,1,2,4,6,7}, 
			            {0,1,3,4,6,7}, {0,1,3,4,5,8}, {0,1,2,4,5,8}, 
			            {0,1,4,5,6,8}, {0,1,2,4,7,8}, {0,1,2,5,7,8}, 
			            {0,1,3,4,7,8}, {0,1,4,5,8,9}, {0,2,3,4,6,8}, 
			            {0,1,2,4,6,8}, {0,2,3,5,6,8}, {0,1,3,4,6,8}, 
			            {0,1,3,5,6,8}, {0,1,3,5,7,8}, {0,1,3,4,6,9}, 
			            {0,1,3,5,6,9}, {0,2,3,6,7,9}, {0,1,3,6,7,9}, 
			            {0,1,4,5,7,9}, {0,2,4,5,7,9}, {0,2,3,5,7,9}, 
			            {0,1,3,5,7,9}, {0,2,4,6,8,10}, {0,1,2,3,4,7}, 
			            {0,1,2,3,4,8}, {0,1,2,3,7,8}, {0,2,3,4,5,8}, 
			            {0,1,2,3,5,8}, {0,1,2,3,6,8}, {0,1,2,3,6,9}, 
			            {0,1,2,5,6,8}, {0,1,2,5,6,9}, {0,2,3,4,6,9}, 
			            {0,1,2,4,6,9}, {0,1,2,4,7,9}, {0,1,2,5,7,9}, 
			            {0,1,3,4,7,9}, {0,1,4,6,7,9}};
    int sixesZ[30] = {36,37,40,41,42,38,46,17,47,44,48,10,49,29,45,3,4,11,12,13,6,24,43,25,19,26,39,28,50,23};

    int sevens[38][7] = {{0,1,2,3,4,5,6}, {0,1,2,3,4,5,7}, {0,1,2,3,4,5,8}, {0,1,2,3,4,6,7},
                         {0,1,2,3,5,6,7}, {0,1,2,3,4,7,8}, {0,1,2,3,6,7,8}, {0,2,3,4,5,6,8},
                         {0,1,2,3,4,6,8}, {0,1,2,3,4,6,9}, {0,1,3,4,5,6,8}, {0,1,2,3,4,7,9},
                         {0,1,2,4,5,6,8}, {0,1,2,3,5,7,8}, {0,1,2,4,6,7,8}, {0,1,2,3,5,6,9},
                         {0,1,2,4,5,6,9}, {0,1,2,3,6,7,10}, {0,1,2,3,6,7,9}, {0,1,2,5,6,7,9},
                         {0,1,2,4,5,8,9}, {0,1,3,4,7,8,9}, {0,2,3,4,5,7,9}, {0,1,2,3,5,7,9},
                         {0,2,3,4,6,7,9}, {0,1,3,4,5,7,9}, {0,1,2,4,5,7,9}, {0,1,3,5,6,7,9},
                         {0,1,2,4,6,7,9}, {0,1,2,4,6,8,9}, {0,1,3,4,6,7,9}, {0,1,3,4,6,8,9},
                         {0,1,3,5,7,9,11},{0,1,3,5,7,9,10}, {0,1,3,5,6,8,10}, {0,1,2,3,5,6,8},
                         {0,1,3,4,5,7,8}, {0,1,2,4,5,7,8}};
    int sevensZ[6] = {36,38,17,12,18,37};

    int eights[29][8] = {{0,1,2,3,4,5,6,7}, {0,1,2,3,4,5,6,8}, {0,1,2,3,4,5,6,9}, {0,1,2,3,4,5,7,8},
                         {0,1,2,3,4,6,7,8}, {0,1,2,3,5,6,7,8}, {0,1,2,3,4,5,8,9}, {0,1,2,3,4,7,8,9}, 
			             {0,1,2,3,6,7,8,9}, {0,2,3,4,5,6,7,9}, {0,1,2,3,4,5,7,9}, {0,1,3,4,5,6,7,9},
                         {0,1,2,3,4,6,7,9}, {0,1,2,4,5,6,7,9}, {0,1,2,3,4,6,8,9}, {0,1,2,3,5,7,8,9},
                         {0,1,3,4,5,6,8,9}, {0,1,2,3,5,6,8,9}, {0,1,2,4,5,6,8,9}, {0,1,2,4,5,7,8,9},
                         {0,1,2,3,4,6,8,10}, {0,1,2,3,5,6,8,10}, {0,1,2,3,5,7,8,10}, {0,1,2,4,5,6,8,10},
                         {0,1,2,4,6,7,8,10}, {0,1,3,4,5,7,8,10}, {0,1,2,4,5,7,8,10}, {0,1,3,4,6,7,9,10},
                         {0,1,2,3,5,6,7,9}}; 
    int eightsZ[2] = {15,29};

    int nines[12][9] = {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,9}, {0,1,2,3,4,5,6,8,9}, 
	                    {0,1,2,3,4,5,7,8,9}, {0,1,2,3,4,6,7,8,9}, {0,1,2,3,4,5,6,8,10}, 
			            {0,1,2,3,4,5,7,8,10}, {0,1,2,3,4,6,7,8,10}, {0,1,2,3,5,6,7,8,10}, 
			            {0,1,2,3,4,6,7,9,10}, {0,1,2,3,5,6,7,9,10}, {0,1,2,4,5,6,8,9,10}};

    if (count > 9 || count < 3)  { /* abort, non-applicable */
        sprintf(forte, "N/A  ");
	    return;
    }
    
    if(count == 3 || count == 9)  {
        members = 12;
    }
    else if(count == 4 || count == 8)  {
	    members = 29;
	    zsets = 2;
    }
    else if(count == 5 || count == 7)  {
	    members = 38;
	    zsets = 6;
    }
    else if(count == 6)  {
	    members = 50;
	    zsets = 29;
    }
	
   
  switch(count)  {
        case 3:
            for(i = 0; i < members; i++)  {
                if (prime[1] == threes[i][1] &&
		            prime[2] == threes[i][2])  {
		            break;
	            }
	        }
            break;

        case 4:
            for(i = 0; i < members; i++)  {
	            if (prime[1] == fours[i][1] &&
		            prime[2] == fours[i][2] &&
		            prime[3] == fours[i][3])  {
		        break;
		        }
	        }
	        for(j = 0; j < zsets; j++)  {
		        if (foursZ[j] == i + 1)  {
	                z = 1;
		            break;
		        }
	        }
	    break;


        case 5:
            for (i = 0; i < members; i++)  {
	            if (prime[1] == fives[i][1] &&
		            prime[2] == fives[i][2] &&
		            prime[3] == fives[i][3] &&
		            prime[4] == fives[i][4])  {
		            break;
		        }
	        }
           
           for(j = 0; j < zsets; j++)  {
	           if (fivesZ[j] == i + 1)  {
	               z = 1;
		           break;
	           }
	       }
	       break;


        case 6:
	        for (i = 0; i < members; i++)  {
		        if (prime[1] == sixes[i][1] &&
		            prime[2] == sixes[i][2] &&
		            prime[3] == sixes[i][3] &&
		            prime[4] == sixes[i][4] &&
		            prime[5] == sixes[i][5])  {
		            break;
		        }
	        }
	        for(j = 0; j < zsets; j++)  {
	            if (sixesZ[j] == i + 1)  {
                    z = 1;
		        break;
                }
            }
	        break;
	    

        case 7:
	        for (i = 0; i < members; i++)  {
		        if (prime[1] == sevens[i][1] &&
		            prime[2] == sevens[i][2] &&
		            prime[3] == sevens[i][3] &&
		            prime[4] == sevens[i][4] &&
		            prime[5] == sevens[i][5] &&
		            prime[6] == sevens[i][6])  {
		            break;
		        }
            }
	        for(j = 0; j < zsets; j++)  {
		        if (sevensZ[j] == i + 1)  {
		            z = 1;
		            break;
		        }
	        }
	        break;

	    
        case 8:
	        for (i = 0; i < members; i++)  {
		        if (prime[1] == eights[i][1] &&
		            prime[2] == eights[i][2] &&
		            prime[3] == eights[i][3] &&
		            prime[4] == eights[i][4] &&
		            prime[5] == eights[i][5] &&
		            prime[6] == eights[i][6] &&
		            prime[7] == eights[i][7])  {
		            break;
		        }
            }
	        for(j = 0; j < zsets; j++)  {
                if (eightsZ[j] == i + 1)  {
	            z = 1;
		        break;
	            }
	        }
	        break;


        case 9:
	        for (i = 0; i < members; i++)  {
		        if (prime[1] == nines[i][1] &&
		            prime[2] == nines[i][2] &&
		            prime[3] == nines[i][3] &&
		            prime[4] == nines[i][4] &&
		            prime[5] == nines[i][5] &&
		            prime[6] == nines[i][6] &&
		            prime[7] == nines[i][7] &&
		            prime[8] == nines[i][8])  {
		           break;
		        }
            }
	        break;
    } /* ends huge switch statement */
    
    
    f2 = ++i;
    if (z == 1)  {  
      theZ = 'Z';
      sprintf(forte, "%d-%c%d", count, theZ, f2);
    }
    else {
      sprintf(forte, "%d-%d", count, f2);
    }
    
    
    return;
    
}



void getICV(long *item, short count, long *icv)  {

    /*  calculate the Interval Class Vector for the incoming set
     *  we do this by measuring the intervals between each member
     *  and all the members greater than it, and keep a tally
     *  of how many of each interval class there is */

    int i = 0, j = 0, k = 0;
    
    /*  zero out the ICV array since it's global in the parent function */
    for(i = 0; i < 6; i++)  {
        icv[i] = 0;
    }

    for (i = 0; i < count - 1; i++)  {
        for (j = i + 1; j < count; j++)  {
	    k = item[j] - item[i];
	    if (k < 0)  {
	        k += 12;
	    }
	    if (k > 6)  {
		k = 12 - k;
            }
	    icv[k-1]++;
        }
    }

    return;

}    




