/* 
 *  main methods for interfacing to max
 */

#include "setcalc.h" 
#include "ext.h"
#include "ext_strings.h"

#define ASSIST_INLET 1
#define ASSIST_OUTLET 2

void *mumble_class;

typedef struct mumble  {
    Object b_ob;
    void *compliment_outlet;
    void *sort_outlet;
    void *normal_outlet;
    void *prime_outlet;
    void *icv_outlet;
    void *forte_outlet;
}  Mumble;


/*  global variables for storing incoming and outgoing data */
t_atom outGoingList[12];
long mySet[100];
long normalform[12];
long primeform[12];
long compliment[12];
long icv[6];
char forte[5];
t_symbol *forteSym;

void *mumble_create(void);
void *mumble_list(Mumble *, t_symbol *, short, t_atom *);
void *mumble_assist(Mumble *, void *, long, long, char *);


int main()  {    
    short i;
    setup((t_messlist**) &mumble_class, (method)mumble_create, 0L, (short)sizeof(Mumble), 0L, 0);
    addmess((method)mumble_list, "list", A_GIMME, 0);
    addmess((method)mumble_assist, "assist", A_CANT, 0);
    
    for(i = 0; i < 12; i++)  {
        outGoingList[i].a_w.w_long = 0;
        outGoingList[i].a_type = A_LONG;
    }
    return 0; 
}



void *mumble_create(void)  {
    
    Mumble *mumble;
    
    /*  create a new instance of this object and return a pointer to it */
    mumble = (Mumble *)newobject(mumble_class);
    
    mumble->compliment_outlet = listout(mumble);
    mumble->sort_outlet = listout(mumble);
    mumble->normal_outlet = listout(mumble);
    mumble->prime_outlet = listout(mumble);
    mumble->icv_outlet = listout(mumble);
    mumble->forte_outlet = outlet_new(mumble, 0);
    
    return(mumble);
    
}


void *mumble_list(Mumble *mumble, t_symbol *msg, short argc, t_atom *argv)  {
    
    /*  this fuction runs when a LIST message is recieved in MAX */
    
    short i;
    short setSize = argc;
    for (i = 0; i < setSize; i++)  {
        switch (argv[i].a_type)  {
            case A_LONG:
                mySet[i] = argv[i].a_w.w_long;
                break;
            case A_SYM:
                break;
            case A_FLOAT:
                break;
        }
        
        
    }
    
    
    /*  here's the actual doing */
    
    // remove duplicate pitch classes, get rid of anything that's not a real pitch class (i.e. 13),
    // and do a fast sort to put things in ascending order.
    setSize = removeDuplicates(mySet, setSize);
    setSize = enforceRange(mySet, setSize);
    shellSort(mySet, setSize);

    short complimentSize = generateCompliment(mySet, setSize, compliment);
    for (i = 0; i < complimentSize; i++)  {  
        outGoingList[i].a_w.w_long = compliment[i];
    }
    outlet_list(mumble->compliment_outlet, 0L, complimentSize, outGoingList);
    
    for (i = 0; i < setSize; i++)  {
        outGoingList[i].a_w.w_long = mySet[i];
    }
    outlet_list(mumble->sort_outlet, 0L, setSize, outGoingList);
    
    findNormalForm(mySet, setSize, normalform);
    for (i = 0; i < setSize; i++)  {
        outGoingList[i].a_w.w_long = normalform[i];
    }
    outlet_list(mumble->normal_outlet, 0L, setSize, outGoingList);
    
    
    findPrimeForm(normalform, setSize, primeform);
    for (i = 0; i < setSize; i++)  {
        outGoingList[i].a_w.w_long = primeform[i];
    }
    outlet_list(mumble->prime_outlet, 0L, setSize, outGoingList);
    
    
    getICV(primeform, setSize, icv);
    for (i = 0; i < 6; i++)  {
        outGoingList[i].a_w.w_long = icv[i];
    }
    outlet_list(mumble->icv_outlet, 0L, 6, outGoingList);
    
    
    getForteNumber(primeform, setSize, forte); 
    forteSym = gensym(forte);
    outlet_anything(mumble->forte_outlet, forteSym, 0, NIL);
    
}


void *mumble_assist(Mumble *mumble, void *b, long msg, long arg, char *dst)  {
    
    /*  this function runs when the user moves the mouse over an inlet or outlet */
    
    if (msg == ASSIST_INLET)  {
        switch (arg)  {
            case 0:
                strcpy(dst, "Input:  List of Pitch Classes");
                break;
        }
    }
    else if (msg == ASSIST_OUTLET)  {
        switch (arg)  {
            case 5:
                strcpy(dst, "List Output:  Complimentary Set.");
                break;
            case 4:
                strcpy(dst, "List Output:  Sorted, Unique Pitches Only.");
                break;
            case 3:
                strcpy(dst, "List Output:  Normal Form.");
                break;
            case 2:
                strcpy(dst, "List Output:  Prime Form.");
                break;
            case 1:
                strcpy(dst, "List Output:  Interval Class Vector.");
                break;
            case 0:
                strcpy(dst, "Symbol Output:  Forte Number.");
                break;
        }
    }
	return 0;
    
}